__author__ = "Vini Salazar"
__license__ = "MIT"
__maintainer__ = "Vini Salazar"
__url__ = "https://github.com/vinisalazar/bioprov"
__version__ = "0.1.9"


"""
Init file for the src/ package.

I removed all modules from here, preferring to import them directly from bioprov/__init__.py
"""
